import logging
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncEngine, async_sessionmaker



logger = logging.getLogger(__name__)


class RepoBase:
    def __init__(self, db: AsyncEngine):
        self.db = db
        self.asmk = async_sessionmaker(self.db, expire_on_commit=False)
        self.curr_session = None
        self._ctx_manager = None
        self.in_transaction = False
    
    
    async def __aenter__(self):
        if self.curr_session is None:
            try:
                self.curr_session = self.asmk()
                self._ctx_manager = self.curr_session.begin()
                await self._ctx_manager.__aenter__()
            except Exception as e:
                logger.error(f"Error starting session: {e}")
                raise
        self.in_transaction = True
        return self
    
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.curr_session is None:
            logger.error("Session is None on __aexit__")
            raise RuntimeError("Session is not initialized before __aexit__")
        
        try:
            await self._ctx_manager.__aexit__(exc_type, exc_val, exc_tb)
            await self.curr_session.close()
        except Exception as e:
            logger.error(f"Error closing session: {e}")
            raise
        finally:
            self._ctx_manager = None
            self.curr_session = None
            self.in_transaction = False
    
    
    async def _commit_if_needed(self):
        if not self.in_transaction and self.curr_session:
            await self.curr_session.commit()
    
    
    async def get_all(self, entity_class, **kwargs):
        async with self.asmk() as sess:
            stmt = select(entity_class).filter_by(**kwargs)
            execres = await sess.execute(stmt)
            return execres.scalars().all()
    
    
    async def get_first(self, entity_class, **kwargs):
        async with self.asmk() as sess:
            stmt = select(entity_class).filter_by(**kwargs)
            execres = await sess.execute(stmt)
            return execres.scalars().first()
    
    
    async def create_if_none(self, ormobj, **kwargs):
        if not self.curr_session:
            raise RuntimeError("No active session. Use 'async with repo:'")
        
        stmt = select(ormobj.__class__).filter_by(**kwargs)
        execres = await self.curr_session.execute(stmt)
        obj = execres.scalars().first()
        if not obj:
            obj = ormobj
            self.curr_session.add(obj)
        return obj
    
    
    async def create(self, *args):
        if not self.curr_session:
            raise RuntimeError("No active session. Use 'async with repo:'")
        for arg in args:
            self.curr_session.merge(arg)
    
    
    async def upsert(self, *args):
        if not self.curr_session:
            raise RuntimeError("No active session. Use 'async with repo:'")
        for arg in args:
            self.curr_session.merge(arg)
    
    
    async def delete(self, *args):
        if not self.curr_session:
            raise RuntimeError("No active session. Use 'async with repo:'")
        for arg in args:
            self.curr_session.delete(arg)



import asyncio
import datetime
import logging
import traceback



class HotAsyncCycler:
    HOT_TIMES = [
    ]
    
    
    def __init__(self, name, func=None,
                 SUPERVISION_SLEEP=1,
                 HOT_SLEEP=0.2,
                 DEFAULT_SLEEP=1,
                 WAS_WORK_SLEEP=1,
                 ERROR_SLEEP=1 * 60):
        self.name = name
        if func is not None:
            self.run1 = func
        self.SUPERVISION_SLEEP = SUPERVISION_SLEEP
        self.HOT_SLEEP = HOT_SLEEP
        self.DEFAULT_SLEEP = DEFAULT_SLEEP
        self.WAS_WORK_SLEEP = WAS_WORK_SLEEP
        self.ERROR_SLEEP = ERROR_SLEEP
        self._current_sleep = self.DEFAULT_SLEEP
        self._last_awake = datetime.datetime.min
    
    
    async def loop1(self):
        logging.info(f"{self.name}: loop1")
        i = 1
        while True:
            logging.debug("Awaked")
            if self.is_default_sleep_passed():
                logging.info(f"{self.name}: Cycle {i}")
                self._last_awake = datetime.datetime.now()
                WAS_WORK = False
                WAS_ERROR = False
                try:
                    WAS_WORK = await self.run1()
                except Exception as e:
                    logging.error(e)
                    logging.error(traceback.format_exc())
                    WAS_ERROR = True
                finally:
                    self._current_sleep = self.refresh_sleeptime(WAS_WORK, WAS_ERROR)
            
            logging.info(f"{self.name}: Cycle {i} completed. Sleeping for {self._current_sleep} sec")
            await asyncio.sleep(self._current_sleep)
            i += 1
    
    
    def refresh_sleeptime(self, was_work, was_error):
        if was_error:
            return self.ERROR_SLEEP
        
        if self.is_hot_now():
            return self.HOT_SLEEP
        
        if was_work:
            return self.WAS_WORK_SLEEP
        
        return self.DEFAULT_SLEEP
    
    
    def is_default_sleep_passed(self):
        return datetime.datetime.now() >= self._last_awake + datetime.timedelta(seconds=self._current_sleep)
    
    
    @staticmethod
    def is_hot_now():
        now_t = datetime.datetime.now().time()
        for start, end in HotAsyncCycler.HOT_TIMES:
            if start <= now_t <= end:
                return True
        return False
    
    
    async def run1(self):
        return False


async def func():
    return False


async def main():
    hac = HotAsyncCycler("2342342", func=func)
    HotAsyncCycler.HOT_TIMES = [(datetime.time(10, 8), datetime.time(10, 9))]
    await hac.loop1()


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG, format="%(asctime)s - %(name)s - %(module)s - %(levelname)s\n%(message)s")
    asyncio.run(main())



"""
print(mapl(lambda x: x + 1, [1, 2, 3]))
print(mapc(set, lambda x: x + 1, [1, 2, 3]))
"""


def mapl(f, it):
    return list(map(f, it))


def filterl(f, it):
    return list(filter(f, it))


def mapc(collect, f, it):
    return collect(map(f, it))


def filterc(collect, f, it):
    return collect(filter(f, it))



import asyncio
import functools
import logging
import time
import traceback
from .utils import to_async



def catch_print_swallow_exc(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except:
            traceback.print_exc()
    
    
    return wrapper


def a_catch_print_swallow_exc(func):
    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except:
            traceback.print_exc()
    
    
    return wrapper


def log_execution_time(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        try:
            result = func(*args, **kwargs)
            return result
        except:
            traceback.print_exc()
        finally:
            end_time = time.perf_counter()
            logging.warning(f"---- {func.__name__} executed in {end_time - start_time:.6f} seconds")
    
    
    return wrapper


def a_log_execution_time(func, level=logging.INFO):
    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        try:
            result = await func(*args, **kwargs)
            return result
        except:
            traceback.print_exc()
        finally:
            end_time = time.perf_counter()
            logging.log(level, f"---- {func.__name__} executed in {end_time - start_time:.6f} seconds")
    
    
    return wrapper


import threading



# Create a thread-local storage object
thread_local = threading.local()
thread_local.call_indent = 0


# F = func
# F(args)
# G = gunc -> Callable
# def gunc(func, args, kwargs):
#     return func(args, kwargs) + 1
# G_F = gunc(func)
#
def log_call(level=logging.INFO, logger=logging.getLogger()):
    def actual_decorator(func):
        if asyncio.iscoroutinefunction(func):
            @functools.wraps(func)
            async def wrapper(*args, **kwargs):
                w = 8
                i = thread_local.call_indent * w
                thread_local.call_indent += 1
                logger.log(level, f"{i * ' '}{func.__name__} Start")
                try:
                    res = await to_async(func, *args, **kwargs)
                    logger.log(level, f"{i * ' '}{func.__name__} Finish")
                    return res
                except:
                    logger.error(f"{i * ' '}{func.__name__} Error")
                    raise
                finally:
                    thread_local.call_indent -= 1
            
            
            return wrapper
        else:
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                logger.log(level, f"{func.__name__} Start")
                try:
                    res = func(*args, **kwargs)
                    logger.log(level, f"{func.__name__} Finish")
                    return res
                except:
                    logger.error(f"{func.__name__} Error")
                    raise
            
            
            return wrapper
    
    
    return actual_decorator



import asyncio
import logging
import time
import traceback



class Cycler:
    def __init__(self, name, async_cycled_func=None, DEFAULT_SLEEP=1, WAS_WORK_SLEEP=1, ERROR_SLEEP=1 * 60):
        self.name = name
        if async_cycled_func is not None:
            self.async_cycled_func = async_cycled_func
        self.DEFAULT_SLEEP = DEFAULT_SLEEP
        self.WAS_WORK_SLEEP = WAS_WORK_SLEEP
        self.ERROR_SLEEP = ERROR_SLEEP
    
    
    def run(self):
        
        i = 1
        while True:
            logging.info(f"{self.name}: Cycle {i}")
            wasWork = False
            sleep_time = self.DEFAULT_SLEEP
            try:
                wasWork = self.cycled_func()
                if wasWork:
                    sleep_time = self.WAS_WORK_SLEEP
            except Exception as e:
                logging.error(e)
                logging.error(traceback.format_exc())
                sleep_time = self.ERROR_SLEEP
            
            logging.info(f"{self.name}: Cycle {i} completed. Sleeping for {sleep_time} sec")
            time.sleep(sleep_time)
            i += 1
    
    
    def cycled_func(self):
        return False


class AsyncCycler:
    def __init__(self, name, async_cycled_func=None, DEFAULT_SLEEP=1, WAS_WORK_SLEEP=1, ERROR_SLEEP=1 * 60):
        self.name = name
        if async_cycled_func is not None:
            self.async_cycled_func = async_cycled_func
        self.DEFAULT_SLEEP = DEFAULT_SLEEP
        self.WAS_WORK_SLEEP = WAS_WORK_SLEEP
        self.ERROR_SLEEP = ERROR_SLEEP
    
    
    async def run(self):
        
        i = 1
        while True:
            logging.info(f"{self.name}: Cycle {i}")
            wasWork = False
            sleep_time = self.DEFAULT_SLEEP
            try:
                wasWork = await self.async_cycled_func()
                if wasWork:
                    sleep_time = self.WAS_WORK_SLEEP
            except Exception as e:
                logging.error(e)
                logging.error(traceback.format_exc())
                sleep_time = self.ERROR_SLEEP
            
            logging.info(f"{self.name}: Cycle {i} completed. Sleeping for {sleep_time} sec")
            await asyncio.sleep(sleep_time)
            i += 1
    
    
    async def async_cycled_func(self):
        return False
